package ch.phoenix.oss.quarkus.commons.tracing;

import io.quarkus.arc.Unremovable;
import io.quarkus.logging.Log;
import io.quarkus.security.identity.SecurityIdentity;
import io.vertx.ext.web.RoutingContext;
import jakarta.inject.Inject;
import jakarta.ws.rs.Path;
import jakarta.ws.rs.container.ContainerRequestContext;
import jakarta.ws.rs.container.ResourceInfo;
import java.util.List;
import org.jboss.resteasy.reactive.server.ServerRequestFilter;

@Unremovable
public class TracingRequestFilter {

    private static final String REDACTED_VALUE = "********";

    private final RoutingContext routingContext;
    private final TracingService tracingService;
    private final SecurityIdentity securityIdentity;
    private final TracingConfiguration configuration;

    @Inject
    public TracingRequestFilter(
            RoutingContext routingContext,
            TracingService tracingService,
            SecurityIdentity securityIdentity,
            TracingConfiguration configuration) {
        this.routingContext = routingContext;
        this.tracingService = tracingService;
        this.securityIdentity = securityIdentity;
        this.configuration = configuration;
    }

    @ServerRequestFilter
    public void filter(ContainerRequestContext requestContext, ResourceInfo resourceInfo) {
        if (securityIdentity.isAnonymous()) {
            tracingService.trace(TracingConstants.ACTOR, TracingConstants.ANONYMOUS);
        } else {
            tracingService.trace(
                    TracingConstants.ACTOR, securityIdentity.getPrincipal().getName());
        }

        var method = requestContext.getMethod();
        tracingService.trace(TracingConstants.REQUEST_METHOD, method);

        var routePattern = getRoutePattern(resourceInfo);
        tracingService.trace(TracingConstants.REQUEST_ROUTE, routePattern);

        var uriInfo = requestContext.getUriInfo();
        uriInfo.getPathParameters()
                .forEach((key, value) ->
                        tracingService.trace(TracingConstants.REQUEST_PATH_PARAMS + '.' + key, joinStrings(value)));

        if (configuration.requestFilter().path().includeRaw()) {
            tracingService.trace(
                    TracingConstants.REQUEST_PATH_RAW, uriInfo.getAbsolutePath().getRawPath());
        }

        var redactedHeaders = configuration
                .requestFilter()
                .headers()
                .redact()
                .orElse(TracingConfiguration.RequestFilterConfiguration.Headers.DEFAULT_REDACTED);

        requestContext.getHeaders().forEach((key, value) -> {
            var lowerCaseKey = key.toLowerCase();
            var property = TracingConstants.REQUEST_HEADERS + '.' + lowerCaseKey;
            if (redactedHeaders.contains(lowerCaseKey)) {
                tracingService.trace(property, REDACTED_VALUE);
            } else {
                tracingService.trace(property, joinStrings(value));
            }
        });

        var redactedQueryParams = configuration
                .requestFilter()
                .query()
                .redact()
                .orElse(TracingConfiguration.RequestFilterConfiguration.Query.DEFAULT_REDACTED);

        uriInfo.getQueryParameters().forEach((key, value) -> {
            var lowerCaseKey = key.toLowerCase();
            var property = TracingConstants.REQUEST_QUERY_PARAMS + '.' + lowerCaseKey;

            if (redactedQueryParams.contains(lowerCaseKey)) {
                tracingService.trace(property, REDACTED_VALUE);
            } else {
                tracingService.trace(property, joinStrings(value));
            }
        });

        if (configuration.requestFilter().query().includeRaw()) {
            var rawQuery = uriInfo.getRequestUri().getRawQuery();
            if (rawQuery != null && !rawQuery.isBlank()) {
                tracingService.trace(TracingConstants.REQUEST_QUERY_RAW, rawQuery);
            }
        }

        tracingService.trace(
                TracingConstants.REQUEST_CLIENT_IP,
                routingContext.request().connection().remoteAddress(true).hostAddress());

        Log.debugf("Incoming request: %s %s", method, uriInfo.getAbsolutePath().getRawPath());
    }

    private static String joinStrings(List<String> value) {
        return String.join(", ", value);
    }

    private String getRoutePattern(ResourceInfo resourceInfo) {
        String classPath = getPathValue(resourceInfo.getResourceClass().getAnnotation(Path.class));
        String methodPath = getPathValue(resourceInfo.getResourceMethod().getAnnotation(Path.class));

        if (!classPath.isEmpty()) {
            if (methodPath.isEmpty()) {
                return "/" + classPath;
            } else {
                return "/" + classPath + "/" + methodPath;
            }
        } else {
            if (methodPath.isEmpty()) {
                return "/";
            } else {
                return "/" + methodPath;
            }
        }
    }

    private static String getPathValue(Path path) {
        if (path == null) {
            return "";
        }
        return trimSlashes(path.value());
    }

    private static String trimSlashes(String segment) {
        if (segment.isEmpty() || "/".equals(segment)) {
            return "";
        }

        // Assuming that it's not possible for a segment to contain //,
        // thus it's possible to avoid extra ifs and whiles, as well
        // as the use of regexes
        int start = (segment.charAt(0) == '/') ? 1 : 0;
        int end = (segment.charAt(segment.length() - 1) == '/') ? segment.length() - 1 : segment.length();

        return segment.substring(start, end);
    }
}
